{ =========================================================================== }
{                                  ALGLIN.PAS                                 }
{ =========================================================================== }

unit AlgLin;

{ ---> Calculo do polinomio caracteristico, polinomio minimo, autovalores e
       autovetores de uma matriz quadrada n x n.

       (C) 1989, 1990 Lenimar Nunes de Andrade
           CCENDM03@BRUFPB.BITNET                                             }

{$E+,N+,F+,R+,B-,V-,S+}

interface

uses
  FPTCBas, Printer, Crt, Matrizes, Equacao;

type
  AutovalorReal = record
    x: real;
    mult: byte;
  end;

  AutovalorComplexo = record
    a, b: real;
    mult: byte;
  end;

  autovalores = record
    QuantRe, QuantComplex: byte;
    AutovalorRe: array[1..MaxM] of AutovalorReal;
    AutovalorCx: array[1..(MaxM div 2)] of AutovalorComplexo;
  end;

  AutovetorReal = record
    dim, n: byte;
    v: array[1..MaxM] of vetor;
  end;

  minimo = record
    expo: array[1..MaxM] of byte;
  end;

procedure PolCaracteristico(mat: matriz; var p: polinomio);
procedure ImprimePolinomioCaracteristico(p: polinomio; letra, variavel: char;
                                                                op: InfoPress);
procedure ImprimePolinomioMinimo(auto: autovalores; letra, variavel: char;
                                                   min: minimo; op: InfoPress);
procedure CalculaAutovalores(p: polinomio; var w: autovalores);
procedure ImprimeAutovalores(w: autovalores; op: InfoPress);
procedure CalculaAutovetores(mat: matriz;lambda: real;var auto: AutovetorReal);
procedure ImprimeAutovetor(auto: AutovetorReal; lambda: real; op: InfoPress);
procedure PolMinimo(mat: matriz; w: autovalores; var min: minimo);
procedure SalvaPolCaracteristico(p: polinomio; tela: TelaPtr);
procedure SalvaAutovalores(w: autovalores; ordem: byte; tela: TelaPtr);

{ =========================================================================== }

implementation

type
  ImpPol = record
    expo, coef: array[1..25*MaxM] of char;
    compr: word;
  end;

const
  erro1: real = 1E-08;   { usado no calculo da multiplicidade dos autovalores }
  erro2: real = 1E-04;   { usado no calculo do polinomio minimo               }

{ --------------------------------------------------------------------------- }

procedure Krylov(mat: matriz; var p: polinomio);

var
  v1, v2: vetor;
  i, j, k, w, ordem: byte;
  solucao: resposta;
  aux: matriz;

begin
  ordem := mat.m;
  aux.m := ordem; aux.n := ordem + 1;
  w := 1;
  repeat
    for k := 1 to ordem do
      if (k = w) then v1[k] := 1 else v1[k] := 0;
    for k := ordem downto 1 do
    begin
      for i := 1 to ordem do
      begin
        v2[i] := 0;
        for j := 1 to ordem do
          v2[i] := v2[i] + mat.a[i, j]*v1[j];
        aux.a[i, k] := v1[i];
      end;
      for j := 1 to ordem do v1[j] := v2[j];
    end;
    for i := 1 to ordem do
      aux.a[i, ordem + 1] := -v2[i];
    ResolveSistema(aux, solucao);
    if (not solucao.SistDet) then Inc(w);
  until ((solucao.SistDet) or (w > mat.m));
  if solucao.SistDet then
  begin
    p.grau := ordem;
    for i := 0 to (ordem - 1) do
    begin
      p.coef[i] := solucao.SolParticular[ordem - i];
      Aproxima(p.coef[i]);
    end;
    p.coef[ordem] := 1;
  end
  else
    p.grau := 0;
end;

{ --------------------------------------------------------------------------- }

function Traco(mat: matriz): real;

var
  soma: real;
  i: byte;

begin
  soma := 0;
  for i := 1 to mat.m do
    soma := soma + mat.a[i, i];
  Traco := soma;
end;

{ --------------------------------------------------------------------------- }

procedure Leverrier(mat: matriz; var p: polinomio);

var
  ordem, i, j, k: byte;
  s, r: vetor;
  soma: real;
  aux: matriz;

begin
  ordem := mat.m;
  p.grau := ordem;
  s[1] := Traco(mat);
  aux := mat;
  for i := 2 to ordem do
  begin
    MultiplicaMatrizes(mat, aux, aux);
    s[i] := Traco(aux);
  end;
  for k := 1 to ordem do
  begin
    soma := 0;
    for i := 1 to k - 1 do
      soma := soma + r[i]*s[k - i];
    r[k] := (-1/k)*(s[k] + soma);
  end;
  p.coef[ordem] := 1;
  for i := 0 to (ordem - 1) do
  begin
    p.coef[i] := r[ordem - i];
    Aproxima(p.coef[i]);
  end;
end;

{ --------------------------------------------------------------------------- }

procedure PolCaracteristico(mat: matriz; var p: polinomio);

begin
  Krylov(mat, p);
  if (p.grau = 0) then Leverrier(mat, p);
end;

{ --------------------------------------------------------------------------- }

procedure FormataPolinomio(p: polinomio; letra, variavel: char; var w: ImpPol;
                                                               racional: byte);

var
  col: word;
  i, k, tam1, tam2: byte;
  x: string;
  sinal: char;

begin
  col := 1; w.expo[col] := ' '; w.coef[col] := letra;
  Inc(col); w.expo[col] := ' '; w.coef[col] := '(';
  Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
  Inc(col); w.expo[col] := ' '; w.coef[col] := ')';
  Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
  Inc(col); w.expo[col] := ' '; w.coef[col] := '=';
  Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
  for i := p.grau downto 0 do
  begin
    if (p.coef[i] <> 0) then
    begin
      FormataElemento(p.coef[i], tam1, tam2);
      if (racional <> 0) then ReformataElemento(p.coef[i], tam1, tam2);
      if i = p.grau then
        if (racional = 0) then
          Str(p.coef[i]:tam1:tam2, x)
        else
          x := Rac(p.coef[i], tam1, tam2)
      else
        if (racional = 0) then
          Str(Abs(p.coef[i]):tam1:tam2, x)
        else
          x := Rac(Abs(p.coef[i]), tam1, tam2);
      if (i < p.grau) then
      begin
        if (p.coef[i] > 0) then sinal := '+' else sinal := '-';
        Inc(col); w.expo[col] := '\'; w.coef[col] := ' ';
        Inc(col); w.expo[col] := '\'; w.coef[col] := sinal;
        Inc(col); w.expo[col] := '\'; w.coef[col] := ' ';
      end;
      if ((Abs(p.coef[i]) <> 1) or (i = 0)) then
        for k := 1 to Length(x) do
        begin
          Inc(col); w.expo[col] := ' '; w.coef[col] := x[k];
        end;
      if (i > 0) then
      begin
        Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
        Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
      end;
      if (i <= 1) then
      begin
        Inc(col); w.coef[col] := ' '; w.expo[col] := ' ';
      end
      else
      begin
        Str(i, x);
        for k := 1 to Length(x) do
        begin
          Inc(col);
          w.coef[col] := ' ';
          w.expo[col] := x[k];
        end;
      end;
    end;
  end;
  Inc(col); w.expo[col] := '#'; w.compr := col - 1;
end;

{ --------------------------------------------------------------------------- }

procedure FormataPolMinimo(auto: autovalores; min: minimo;
                         letra, variavel: char; var w: ImpPol; racional: byte);

var
  col: word;
  i, k, tam1, tam2: byte;
  x: string;
  sinal: char;

begin
  col := 1; w.expo[col] := ' '; w.coef[col] := letra;
  Inc(col); w.expo[col] := ' '; w.coef[col] := '(';
  Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
  Inc(col); w.expo[col] := ' '; w.coef[col] := ')';
  Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
  Inc(col); w.expo[col] := ' '; w.coef[col] := '=';
  Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
  for i := 1 to auto.QuantRe do
  begin
    Inc(col); w.expo[col] := '\'; w.coef[col] := '(';
    Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
    if (auto.AutovalorRe[i].x <> 0) then
    begin
      FormataElemento(auto.AutovalorRe[i].x, tam1, tam2);
      if (racional <> 0) then
        ReformataElemento(auto.AutovalorRe[i].x, tam1, tam2);
      if (auto.AutovalorRe[i].x > 0) then sinal := '-' else sinal := '+';
      if (racional = 0) then
        Str(Abs(auto.AutovalorRe[i].x):tam1:tam2, x)
      else
        x := Rac(Abs(auto.AutovalorRe[i].x), tam1, tam2);
      Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
      Inc(col); w.expo[col] := ' '; w.coef[col] := sinal;
      Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
      for k := 1 to Length(x) do
      begin
        Inc(col); w.expo[col] := ' '; w.coef[col] := x[k];
      end;
    end;
    Inc(col); w.expo[col] := ' '; w.coef[col] := ')';
    if (min.expo[i] > 1) then
    begin
      Str(min.expo[i], x);
      for k := 1 to Length(x) do
      begin
        Inc(col); w.expo[col] := x[k]; w.coef[col] := ' ';
      end;
    end;
  end;
  for i := 1 to auto.QuantComplex do
  begin
    Inc(col); w.expo[col] := '\'; w.coef[col] := '(';
    Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
    Inc(col); w.expo[col] := '2'; w.coef[col] := ' ';
    Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
    if (auto.AutovalorCx[i].a <> 0) then
    begin
      FormataElemento(2*auto.AutovalorCx[i].a, tam1, tam2);
      if (racional <> 0) then
        ReformataElemento(2*auto.AutovalorCx[i].a, tam1, tam2);
      if (auto.AutovalorCx[i].a > 0) then sinal := '-' else sinal := '+';
      Inc(col); w.expo[col] := ' '; w.coef[col] := sinal;
      if (Abs(2*auto.AutovalorCx[i].a) <> 1) then
      begin
        if (racional = 0) then
          Str(Abs(2*auto.AutovalorCx[i].a):tam1:tam2, x)
        else
          x := Rac(Abs(2*auto.AutovalorCx[i].a), tam1, tam2);
        Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
        for k := 1 to Length(x) do
        begin
          Inc(col); w.expo[col] := ' '; w.coef[col] := x[k];
        end;
      end;
      Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
      Inc(col); w.expo[col] := ' '; w.coef[col] := variavel;
      Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
    end;
    Inc(col); w.expo[col] := ' '; w.coef[col] := '+';
    Inc(col); w.expo[col] := ' '; w.coef[col] := ' ';
    FormataElemento(Sqr(auto.AutovalorCx[i].a) + Sqr(auto.AutovalorCx[i].b),
                                                                   tam1, tam2);
    if (racional <> 0) then
      ReformataElemento(Sqr(auto.AutovalorCx[i].a)+Sqr(auto.AutovalorCx[i].b),
                                                                   tam1, tam2);
    if (racional = 0) then
      Str((Sqr(auto.AutovalorCx[i].a) + Sqr(auto.AutovalorCx[i].b)):tam1:tam2,
                                                                            x)
    else
      x := Rac((Sqr(auto.AutovalorCx[i].a) + Sqr(auto.AutovalorCx[i].b)),
                                                                   tam1, tam2);
    for k := 1 to Length(x) do
    begin
      Inc(col); w.expo[col] := ' '; w.coef[col] := x[k];
    end;
    Inc(col); w.expo[col] := ' '; w.coef[col] := ')';
    if (min.expo[i + auto.QuantRe] > 1) then
    begin
      Str(min.expo[i + auto.QuantRe], x);
      for k := 1 to Length(x) do
      begin
        Inc(col); w.expo[col] := x[k]; w.coef[col] := ' ';
      end;
    end;
  end;
  Inc(col); w.expo[col] := '#'; w.compr := col - 1;
end;

{ --------------------------------------------------------------------------- }

procedure ImprimePolinomio(alc, largura, QuantLinhas, opcao1, opcao2: byte;
                                                   w: ImpPol; opbrancos: byte);

var
  LinhasImpressas, caracter, brancos, corte, i, j, k, col: word;
  fim: boolean;
  ch: char;

begin
  LinhasImpressas := 0;
  fim := false; corte := 0; k := 1; caracter := 1; brancos := 0;
  repeat
    j := caracter;
    while (j - caracter + 1 + brancos < largura) and (w.expo[j] <> '#') do
    begin
      Inc(j);
      if w.expo[j] = '\' then
        if (opcao2 = 0) then corte := j else corte := j - 1;
    end;
    if (w.expo[j] = '#') then
    begin
      corte := j - 1;
      fim := true;
    end;
    if (LinhasImpressas > 0) then
      if (opcao1 = 1) then
        Write(Spc(opbrancos) + '       ')
      else
        Write(Lst, Spc(opbrancos) + '       ')
    else
      if (opcao1 = 1) then
        Write(Spc(opbrancos))
      else
        Write(Lst, Spc(opbrancos));
    for i := k to corte do
      if (w.expo[i] <> '\') then
        if (opcao1 = 1) then Write(w.expo[i]) else Write(Lst, w.expo[i])
      else
        if (opcao1 = 1) then Write(' ') else Write(Lst, ' ');
    if (opcao1 = 1) then Writeln else Writeln(Lst);
    if (LinhasImpressas > 0) then
      if (opcao1 = 1) then
        Write(Spc(opbrancos) + '       ')
      else
        Write(Lst, Spc(opbrancos) + '       ')
    else
      if (opcao1 = 1) then
        Write(Spc(opbrancos))
      else
        Write(Lst, Spc(opbrancos));
    for i := k to corte do
      if (opcao1 = 1) then Write(w.coef[i]) else Write(Lst, w.coef[i]);
    if (opcao1 = 1) then
    begin
      if (not fim) then
      begin
        Writeln; Writeln;
      end
    end
    else
    begin
      Writeln(Lst); Writeln(Lst);
    end;
    brancos := 7;

    Inc(LinhasImpressas);

    if (LinhasImpressas mod QuantLinhas = 0) and (opcao1 = 1)
                                                             and (not fim) then
    begin
      if (alc < 1) then
      begin
        Writeln;
        Write('---> Pressione qualquer tecla para continuar');
        ch := ReadKey;
        Writeln; Writeln;
      end
      else
      begin
        if (alc = 1) then
          PausaJanMenu(' Polinomio caracteristico ')
        else
          PausaJanMenu(' Polinomio minimo ');
      end;
    end;
    k := corte + 1; caracter := corte + 1;
  until fim;
  if (alc <> 1) then
    if (opcao1 = 1) then Writeln else Writeln(Lst);
end;

{ --------------------------------------------------------------------------- }

procedure ImprimePolinomioCaracteristico(p: polinomio; letra, variavel: char;
                                                                op: InfoPress);

var
  w: ImpPol;

begin
  op.brancos := 0;
  FormataPolinomio(p, letra, variavel, w, op.racional);
  if (w.compr < op.largura) then
  begin
    Writeln;
    op.brancos := (op.largura - w.compr) div 2;
  end;
  ImprimePolinomio(op.opcao, op.largura, op.MaxL, op.tela, 0, w, op.brancos);
end;

{ --------------------------------------------------------------------------- }

procedure ImprimePolinomioMinimo(auto: autovalores; letra, variavel: char;
                                                   min: minimo; op: InfoPress);

var
  w: ImpPol;
  i: byte;

begin
  op.brancos := 0;
  if (op.opcao = 1) then
  begin
    for i := 1 to auto.QuantRe do min.expo[i] := auto.AutovalorRe[i].mult;
    for i := 1 to auto.QuantComplex do
      min.expo[i + auto.QuantRe] := auto.AutovalorCx[i].mult;
  end;
  FormataPolMinimo(auto, min, letra, variavel, w, op.racional);
  if (w.compr < op.largura) then
  begin
    Writeln;
    op.brancos := (op.largura - w.compr) div 2;
  end;
  if (op.opcao >= 1) then
    if (op.opcao = 1) then
      ImprimePolinomio(1, op.largura, op.MaxL, op.tela, 1, w, op.brancos)
    else
      ImprimePolinomio(2, op.largura, op.MaxL, op.tela, 1, w, op.brancos)
  else
    ImprimePolinomio(0, op.largura, op.MaxL, op.tela, 1, w, op.brancos);
end;

{ --------------------------------------------------------------------------- }

procedure CalculaAutovalores(p: polinomio; var w: autovalores);

var
  i, j: byte;
  resp: solucoes;
  novo: boolean;

begin
  ResolveEquacao(p, 0, resp);
  if (not resp.resolveu) then ResolveEquacao(p, 1, resp);
  for i := 1 to p.grau do
  begin
    Aproxima(resp.x[i].Re);
    Aproxima(resp.x[i].Im);
  end;
  w.QuantRe := 0; w.QuantComplex := 0;
  for i := 1 to p.grau do
  begin
    if (resp.x[i].Im = 0) then
    begin
      novo := true;
      if (w.QuantRe = 0) then
      begin
        w.QuantRe := 1;
        w.AutovalorRe[1].x := resp.x[i].Re;
        w.AutovalorRe[1].mult := 1;
      end
      else
      begin
        for j := 1 to w.QuantRe do
          if (Abs(resp.x[i].Re - w.AutovalorRe[j].x) <= erro1) then
          begin
            novo := false;
            Inc(w.AutovalorRe[j].mult);
          end;
          if novo then
          begin
            Inc(w.QuantRe);
            w.AutovalorRe[w.QuantRe].x := resp.x[i].Re;
            w.AutovalorRe[w.QuantRe].mult := 1;
          end;
      end;
    end
    else
    begin
      if (resp.x[i].Im > 0) then
      begin
        novo := true;
        if (w.QuantComplex = 0) then
        begin
          w.QuantComplex := 1;
          w.AutovalorCx[1].a := resp.x[i].Re;
          w.AutovalorCx[1].b := resp.x[i].Im;
          w.AutovalorCx[1].mult := 1;
        end
        else
        begin
          for j := 1 to w.QuantComplex  do
            if (Abs(resp.x[i].Re - w.AutovalorCx[j].a) <= erro1) and
                         (Abs(resp.x[i].Im - w.AutovalorCx[j].b) <= erro1) then
            begin
              novo := false;
              Inc(w.AutovalorCx[j].mult);
            end;
            if novo then
            begin
              Inc(w.QuantComplex);
              w.AutovalorCx[w.QuantComplex].a := resp.x[i].Re;
              w.AutovalorCx[w.QuantComplex].b := resp.x[i].Im;
              w.AutovalorCx[w.QuantComplex].mult := 1;
            end;
        end;
      end;
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure ImprimeAutovalores(w: autovalores; op: InfoPress);

var
  QuantImpress, i, j, tam1, tam2, tam3, tam4: byte;
  fim: boolean;

const
  brancos: byte = 10;

procedure Verifica(var x: byte; var fim: boolean);

begin { ---> Verifica }
  Inc(x);
  if (x = op.n) then fim := true else fim := false;
  if ((x mod 6 = 0) and (not fim) and (op.opcao = 1)) then
    PausaJanMenu(' Autovalores ');
end; { ---> Verifica }

begin { ---> ImprimeAutovalores }
  QuantImpress := 0; fim := false;
  if (op.opcao = 1) then brancos := 20;
  if (w.QuantRe > 0) then
  begin
    if (op.tela = 1) and (op.opcao <> 1) then
    begin
      Writeln('Autovalores reais: '); Writeln;
    end
    else
      if (op.tela = 0) then
      begin
        Writeln(Lst, 'Autovalores reais: '); Writeln(Lst);
      end;
    for i := 1 to w.QuantRe do
    begin
      FormataElemento(w.AutovalorRe[i].x, tam1, tam2);
      if (op.racional <> 0) then
        ReformataElemento(w.AutovalorRe[i].x, tam1, tam2);
      for j := 1 to w.AutovalorRe[i].mult do
      begin
        if (op.tela = 1) then
          if (op.racional = 0) then
            Writeln(' ':brancos, w.AutovalorRe[i].x:tam1:tam2)
          else
            Writeln(' ':brancos, Rac(w.AutovalorRe[i].x, tam1, tam2))
        else
          if (op.racional = 0) then
            Writeln(Lst, ' ':brancos, w.AutovalorRe[i].x:tam1:tam2)
          else
            Writeln(Lst, ' ':brancos, Rac(w.AutovalorRe[i].x, tam1, tam2));
        Verifica(QuantImpress, fim);
      end;
    end;
    if (op.tela = 1) and (op.opcao <> 1) then
      Writeln
    else
      if (op.tela = 0) then Writeln(Lst);
  end;
  if (w.QuantComplex > 0) then
  begin
    if (op.tela = 1) and (op.opcao <> 1) then
    begin
      Writeln('Autovalores complexos:'); Writeln;
    end
    else
      if (op.tela = 0) then
      begin
        Writeln(Lst, 'Autovalores complexos:'); Writeln(Lst);
      end;
    for i := 1 to w.QuantComplex do
    begin
      FormataElemento(w.AutovalorCx[i].a, tam1, tam2);
      if (op.racional <> 0) then
        ReformataElemento(w.AutovalorCx[i].a, tam1, tam2);
      FormataElemento(w.AutovalorCx[i].b, tam3, tam4);
      if (op.racional <> 0) then
        ReformataElemento(w.AutovalorCx[i].b, tam3, tam4);
      for j := 1 to w.AutovalorCx[i].mult do
      begin
        if (op.tela = 1) then
        begin
          if (op.racional = 0) then
          begin
            Write(' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(w.AutovalorCx[i].a:tam1:tam2, ' + ')
            else
              Write(' ':(tam1 + 3));
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(w.AutovalorCx[i].b:tam3:tam4, ' i')
            else
              Writeln(' ':tam1, ' i');
          end
          else
          begin
            Write(' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Rac(w.AutovalorCx[i].a, tam1, tam2), ' + ')
            else
              Write(' ':(tam1 + 3));
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(Rac(w.AutovalorCx[i].b, tam3, tam4), ' i')
            else
              Writeln(' ':tam1, ' i');
          end;
          Verifica(QuantImpress, fim);
          if (op.racional = 0) then
          begin
            Write(' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(w.AutovalorCx[i].a:tam1:tam2)
            else
              Write(' ':tam1);
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(' - ', w.AutovalorCx[i].b:tam3:tam4, ' i')
            else
              Writeln(' ':(tam1 + 2), '- i');
          end
          else
          begin
            Write(' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Rac(w.AutovalorCx[i].a, tam1, tam2))
            else
              Write(' ':tam1);
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(' - ', Rac(w.AutovalorCx[i].b, tam3, tam4), ' i')
            else
              Writeln(' ':(tam1 + 2), '- i');
          end;
        end
        else
        begin
          if (op.racional = 0) then
          begin
            Write(Lst, ' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Lst, w.AutovalorCx[i].a:tam1:tam2, ' + ')
            else
              Write(Lst, ' ':(tam1 + 3));
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(Lst, w.AutovalorCx[i].b:tam3:tam4, ' i')
            else
              Writeln(Lst, ' ':tam1, ' i');
          end
          else
          begin
            Write(Lst, ' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Lst, Rac(w.AutovalorCx[i].a, tam1, tam2), ' + ')
            else
              Write(Lst, ' ':(tam1 + 3));
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(Lst, Rac(w.AutovalorCx[i].b, tam3, tam4), ' i')
            else
              Writeln(Lst, ' ':tam1, ' i');
          end;
          if (op.racional = 0) then
          begin
            Write(Lst, ' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Lst, w.AutovalorCx[i].a:tam1:tam2)
            else
              Write(Lst, ' ':tam1);
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(Lst, ' - ', w.AutovalorCx[i].b:tam3:tam4, ' i')
            else
              Writeln(Lst, ' ':(tam1 + 2), '- i');
          end
          else
          begin
            Write(Lst, ' ':brancos);
            if (w.AutovalorCx[i].a <> 0) then
              Write(Lst, Rac(w.AutovalorCx[i].a, tam1, tam2))
            else
              Write(Lst, ' ':tam1);
            if (w.AutovalorCx[i].b <> 1) then
              Writeln(Lst, ' - ', Rac(w.AutovalorCx[i].b, tam3, tam4), ' i')
            else
              Writeln(Lst, ' ':(tam1 + 2), '- i');
          end;
        end;
        Verifica(QuantImpress, fim);
      end;
    end;
  end;
  if (op.tela = 1) and (op.opcao <> 1) then
    Writeln
  else
    if (op.tela = 0) then Writeln(Lst);
end; { ---> ImprimeAutovalores }

{ --------------------------------------------------------------------------- }

procedure CalculaAutovetores(mat: matriz; lambda: real;
                                                      var auto: AutovetorReal);

var
  i, j: byte;
  aux: matriz;
  solucao: resposta;

begin
  auto.n := mat.n;
  aux.m := mat.n; aux.n := mat.n + 1;
  for j := 1 to aux.m do aux.a[j, aux.n] := 0;
  for i := 1 to mat.m do
    for j := 1 to mat.n do
      if (i = j) then
        aux.a[i, j] := mat.a[i, j] - lambda
      else
        aux.a[i, j] := mat.a[i, j];
  ResolveSistema(aux, solucao);
  auto.dim := solucao.x.m;
  for i := 1 to auto.dim do
    for j := 1 to mat.n do
      auto.v[i][j] := solucao.x.a[i, j];
end;

{ --------------------------------------------------------------------------- }

procedure ImprimeAutovetor(auto: AutovetorReal; lambda: real; op: InfoPress);

var
  i, j, k, tam1, tam2, imp, LinhasImpressas: byte;
  aux: matriz;
  ch: char;
  fim: boolean;

procedure PausaAutovet(fim: boolean);

var
  c: char;

begin { ---> PausaAutovet }
  Cor(7, 0);
  if (op.n > 5) then
    Window(11, 11, 70, 20)
  else
    Window(11, 22, 70, 22);
  if fim then
  begin
    if (op.n > 5) then
    begin
      GoToXY(1, 10); Write(#192, Repete(57, #196), #217);
    end
    else
      Write('   F3: Nova matriz    F4: Formato racional   F7: Volta    ');
  end
  else
  begin
    if (op.n > 5) then
    begin
      GoToXY(10, 10);
      Write(' Pressione qualquer tecla para continuar ');
    end
    else
      Write('           Pressione qualquer tecla para continuar       ');
    repeat
      c := ReadKey;
    until (c <> #0);
    if not fim then
    begin
      if (op.n > 5) then
        Window(12, 14, 68, 19)
      else
      begin
        Cor(2, 7);
        Window(11, 17, 68, 19);
      end;
      ClrScr;
      if (op.n <= 12) then Writeln;
    end;
  end;
end; { ---> PausaAutovet }

const
  brancos = 10;

begin { ---> ImprimeAutovetor }
  fim := false;
  FormataElemento(lambda, tam1, tam2);
  if (op.racional <> 0) then ReformataElemento(lambda, tam1, tam2);
  if (op.opcao = 1) then
  begin
     if (op.n > 5) then
     begin
       Window(12, 12, 68, 19); Cor(7, 0);
     end
     else
     begin
       Window(12, 14, 68, 19); Cor(2, 7);
     end;
     ClrScr;
  end;
  if (op.tela = 1) then
  begin
    Writeln;
    if (op.racional = 0) then
      Writeln(' [ ', #224, ' = ', lambda:tam1:tam2, ' * dim = ',
                                                            auto.dim, ' ] ')
    else
      Writeln(' [ ', #224, ' = ', Rac(lambda, tam1, tam2), ' * dim = ',
                                                             auto.dim,' ] ');
    Writeln;
  end
  else
  begin
    Writeln(Lst);
    if (op.racional = 0) then
      Writeln(Lst, '[ x = ', lambda:tam1:tam2, ' * dim = ', auto.dim, ' ] ')
    else
      Writeln(Lst, '[ x = ', Rac(lambda, tam1, tam2), ' * dim = ', auto.dim,
                                                                       ' ] ');
    Writeln(Lst);
  end;
  aux.m := auto.dim; aux.n := auto.n;
  for j := 1 to aux.m do
    for k := 1 to aux.n do
      aux.a[j, k] := auto.v[j, k];
  FormataMatriz(aux, tam1, tam2);
  if (op.racional <> 0) then ReformataMatriz(aux, tam1, tam2);
  if (op.opcao = 1) then
  begin
    if (op.n > 5) then
      Window(12, 14, 68, 19)
    else
      Window(12, 17, 68, 19);
    ClrScr;
    if (op.n <= 12) then Writeln;
  end;
  for j := 1 to auto.dim do
  begin
    if (op.tela = 1) then
      Write(' ':brancos, '(')
    else
      Write(Lst, ' ':brancos, '(');
    imp := brancos + 1; LinhasImpressas := 0;
    for k := 1 to auto.n do
    begin
      if (op.tela = 1) then
        if (imp + tam1 + 1 <= op.largura) then
        begin
          if (op.racional = 0) then
            Write(auto.v[j, k]:tam1:tam2)
          else
            Write(Rac(auto.v[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
        else
        begin
          imp := brancos + 1;
          Writeln; Inc(LinhasImpressas);
          if (LinhasImpressas mod op.MaxL = 0) and (op.tela = 1) then
          begin
            Write('Pressione qualquer tecla para continuar.');
            ch := ReadKey;
          end;
          if (op.racional = 0) then
            Write(' ':imp, auto.v[j, k]:tam1:tam2)
          else
            Write(' ':imp, Rac(auto.v[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
      else
        if (imp + tam1 + 1 <= op.largura) then
        begin
          if (op.racional = 0) then
            Write(Lst, auto.v[j, k]:tam1:tam2)
          else
            Write(Lst, Rac(auto.v[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end
        else
        begin
          imp := brancos + 1;
          Writeln(Lst);
          if (op.racional = 0) then
            Write(Lst, ' ':imp, auto.v[j, k]:tam1:tam2)
          else
            Write(Lst, ' ':imp, Rac(auto.v[j, k], tam1, tam2));
          imp := imp + tam1 + 1;
        end;
      if (k < auto.n) then
        if (op.tela = 1) then Write(',') else Write(Lst, ',');
    end;
    if (op.tela = 1) then Writeln(')') else Writeln(Lst, ')');
    fim := op.ultimo and (j = auto.dim);
    if (op.opcao = 1) and (op.tela = 1) then PausaAutovet(fim);
  end;
  if (op.tela <> 1) then Writeln(Lst);
end; { ---> ImprimeAutovetor }

{ --------------------------------------------------------------------------- }

procedure TestaMNula(mat: matriz; aux: minimo; w: autovalores;
                                                            var nula: boolean);

var
  ordem, i, j, k: byte;
  m1, m2, MAux: matriz;

begin
  nula := true; ordem := mat.m;
  MAux.m := ordem; MAux.n := ordem;
  m1.m := ordem; m1.n := ordem;
  m2.m := ordem; m2.n := ordem;
  for i := 1 to ordem do
    for j := 1 to ordem do
      if (i = j) then MAux.a[i, j] := 1 else MAux.a[i, j] := 0;
  for k := 1 to w.QuantRe do
  begin
    for i := 1 to ordem do
      for j := 1 to ordem do
        if (i = j) then
          m1.a[i, j] := mat.a[i, j] - w.AutovalorRe[k].x
        else
          m1.a[i, j] := mat.a[i, j];
    for i := 1 to aux.expo[k] do MultiplicaMatrizes(MAux, m1, MAux);
  end;
  MultiplicaMatrizes(mat, mat, m2);
  for k := 1 to w.QuantComplex do
  begin
    for i := 1 to ordem do
      for j := 1 to ordem do
        if (i = j) then
          m1.a[i, j] := m2.a[i, j] - 2*w.AutovalorCx[k].a*mat.a[i, j] +
                              Sqr(w.AutovalorCx[k].a) + Sqr(w.AutovalorCx[k].b)
        else
          m1.a[i, j] := m2.a[i, j] - 2*w.AutovalorCx[k].a*mat.a[i, j];
    for i := 1 to aux.expo[k + w.QuantRe] do MultiplicaMatrizes(MAux, m1,MAux);
  end;
  for i := 1 to ordem do
    for j := 1 to ordem do
      if (Abs(MAux.a[i, j]) > erro2) then nula := false;
end;

{ --------------------------------------------------------------------------- }

procedure PolMinimo(mat: matriz; w: autovalores; var min: minimo);

var
  q, i, k: byte;
  max, aux: minimo;
  ProximoK, MatrizNula: boolean;

begin
  q := w.QuantRe + w.QuantComplex;
  for i := 1 to w.QuantRe do
    max.expo[i] := w.AutovalorRe[i].mult;
  for i := 1 to w.QuantComplex do
    max.expo[i + w.QuantRe] := w.AutovalorCx[i].mult;
  for i := 1 to q do min.expo[i] := 1;
  TestaMNula(mat, min, w, MatrizNula);
  if MatrizNula then Exit;
  aux := max;
  for k := 1 to q do
    if (aux.expo[k] > min.expo[k]) then
    begin
      ProximoK := false;
      while (not ProximoK) do
      begin
        TestaMNula(mat, aux, w, MatrizNula);
        if MatrizNula then
          if (aux.expo[k] > min.expo[k]) then
            Dec(aux.expo[k])
          else
            ProximoK := true
        else
        begin
          min.expo[k] := aux.expo[k] + 1;
          aux.expo[k] := min.expo[k];
          ProximoK := true;
        end;
      end;
    end;
end;

{ --------------------------------------------------------------------------- }

procedure SalvaPolCaracteristico(p: polinomio; tela: TelaPtr);

var
  arq: text;
  i: byte;
  HouveErro: boolean;

begin
  Assign(arq, 'POLINOM.CAR');
  {$I-}
  Rewrite(arq);
  {$I+}
  if (IOResult <> 0) then
  begin
    RestauraTela(tela);
    if (p.grau <= 5) then Erro(1, 2, HouveErro) else Erro(1, 1, HouveErro);
    Exit;
  end;
  Writeln(arq); Writeln(arq, '--------------------------------------');
  Writeln(arq);
  Writeln(arq, 'Grau do polinomio: ', p.grau);
  Writeln(arq);
  Writeln(arq, 'Coeficientes em ordem decrescente das');
  Writeln(arq, 'potencias de ''x'':');
  for i := p.grau downto 0 do
  begin
    Writeln(arq);
    Writeln(arq, p.coef[i]:30:20);
  end;
  Writeln(arq); Writeln(arq, '--------------------------------------');
  Close(arq);
end;

{ --------------------------------------------------------------------------- }

procedure SalvaAutovalores(w: autovalores; ordem: byte; tela: TelaPtr);

var
  arq: text;
  i, j: byte;
  HouveErro: boolean;

begin
  Assign(arq, 'AUTOVALO.RES');
  {$I-}
  Rewrite(arq);
  {$I+}
  if (IOResult <> 0) then
  begin
    RestauraTela(tela);
    if (ordem <= 5) then Erro(1, 2, HouveErro) else Erro(1, 1, HouveErro);
    Exit;
  end;
  Writeln(arq); Writeln(arq,
          '-----------------------------------------------------------------');
  Writeln(arq);
  if (w.QuantRe > 0) then
  begin
    Writeln(arq, 'Autovalores reais: '); Writeln(arq);
    for i := 1 to w.QuantRe do
      for j := 1 to w.AutovalorRe[i].mult do
        Writeln(arq, w.AutovalorRe[i].x:30:20);
    Writeln(arq);
  end;
  if (w.QuantComplex > 0) then
  begin
    Writeln(arq, 'Autovalores complexos: '); Writeln(arq);
    for i := 1 to w.QuantComplex do
      for j := 1 to w.AutovalorCx[i].mult do
        Writeln(arq, w.AutovalorCx[i].a:30:20, '  ', #241,
                                               w.AutovalorCx[i].b:30:20, ' i');
  end;
  Writeln(arq);
  Writeln(arq,
          '-----------------------------------------------------------------');
  Close(arq);
end;

{ --------------------------------------------------------------------------- }

var
  arq: file of real;

begin
  Assign(arq, 'ALGLIN.ERR');
  {$I-}
  Reset(arq);
  {$I+}
  if (IOResult = 0) then
  begin
    Read(arq, erro1, erro2);
    Close(arq);
  end;
end.

{ =========================================================================== }

