{ =========================================================================== }
{                                 EQUACAO.PAS                                 }
{ =========================================================================== }

{$E+,N+,F+,R+,B-,S+}

unit Equacao;

{ ---> Resolucao de equacoes algebricas

       (C) 1988, 1990 Lenimar Nunes de Andrade
           CCENDM03@BRUFPB.BITNET                                             }

interface

const
  GrauMax = 30;  { maior grau admissivel para a equacao }

type
  real = extended;

  polinomio = record
    coef: array[0..GrauMax + 1] of real;
    grau: byte;
  end;

  complexo = record
    Re, Im: real;
    metodo: string[5];
  end;

  solucoes = record
    x: array[1..GrauMax] of complexo;
    resolveu: boolean;
  end;

function Fz(f: polinomio; a, b: real): real;
procedure ResolveEquacao(a: polinomio; op: byte; var resp: solucoes);

{ =========================================================================== }

implementation

type
  PolInteiro = record
    a: array[0..GrauMax] of longint;
    quant: byte;
  end;

const
  erro1: real = 1E-10; { usado na resolucao de equacoes do 2o. grau }
                       {                               2            }
  erro2: real = 1E-10; { usado na divisao de p(x) por x  + bx + c   }
  erro3: real = 1E-10; { usado na divisao de p(x) por x - p/q       }
  erro4: real = 1E-06; { usado na aproximacao para inteiros         }
  erro5: real = 1E-16; { usado no Algoritmo de Bairstow             }
  erro6: real = 1E-10; { usado no Algoritmo de Bairstow             }

{ --------------------------------------------------------------------------- }

function Min(a, b: longint): longint;

begin
  if (a < b) then Min := a else Min := b;
end;

{ --------------------------------------------------------------------------- }

function Fz(f: polinomio; a, b: real): real;

{ ---> calcula |f(a + bi)|                                                    }

var
  k: byte;
  g: polinomio;
  p, q, r1, r2: real;

begin
  if (b <> 0) then
  begin
    p := -2*a; q := a*a + b*b;
    g.coef[f.grau] := 0; g.coef[f.grau - 1] := 0;
    for k := (f.grau - 2) downto 0 do
      g.coef[k] := f.coef[k + 2] - p*g.coef[k + 1] - q*g.coef[k + 2];
    r1 := f.coef[1] - g.coef[1]*q - g.coef[0]*p;
    r2 := f.coef[0] - g.coef[0]*q;
    Fz := Sqrt(Sqr(r1) + Sqr(r2))
  end
  else
  begin
    g.coef[f.grau + 1] := 0;
    for k := f.grau downto 0 do
      g.coef[k] := f.coef[k] + g.coef[k + 1]*a;
    Fz := Abs(g.coef[0])
  end;
end;

{ --------------------------------------------------------------------------- }

procedure SegGrau(b, c: real; var resp: solucoes; var sol: byte);

{                         2
  ---> Resolve a equacao x  + b x + c = 0                                     }

var
  delta: real;

begin
  delta := Sqr(b) - 4*c;
  if (Abs(delta) <= erro1) then delta := 0;
  if (delta >= 0) then
  begin
    sol := sol + 1;
    resp.x[sol].Re := (-b + Sqrt(delta))/2; resp.x[sol].Im := 0;
    sol := sol + 1;
    resp.x[sol].Re := (-b - Sqrt(delta))/2; resp.x[sol].Im := 0;
  end
  else
  begin
    sol := sol + 1;
    resp.x[sol].Re := -b/2; resp.x[sol].Im := Sqrt(-delta)/2;
    sol := sol + 1;
    resp.x[sol].Re := -b/2; resp.x[sol].Im := -Sqrt(-delta)/2;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure DivPolinomio(var c: polinomio; p, q: real; var DivExata: boolean;
                       var resp: solucoes; var sol: byte);

{                               n                          2
  ---> Divide o polinomio c[n] x + ... + c[1]x + c[0] por x  + px + q         }

var
  i, k: byte;
  g: polinomio;

begin
  g.coef[c.grau] := 0; g.coef[c.grau - 1] := 0;
  for k := (c.grau - 2) downto 0 do
    g.coef[k] := c.coef[k + 2] - p*g.coef[k + 1] - q*g.coef[k + 2];
  DivExata := (Abs(g.coef[1]*q + g.coef[0]*p - c.coef[1]) <= erro2)
                                   and (Abs(g.coef[0]*q - c.coef[0]) <= erro2);
  if DivExata then
  begin
    c.grau := c.grau - 2;
    for i := 0 to c.grau do c.coef[i] := g.coef[i];
    SegGrau(p, q, resp, sol);
  end;
end;

{ --------------------------------------------------------------------------- }

procedure DividePol(var c: polinomio; p, q: longint; var resp: solucoes;
                    var sol: byte; metodo: string);

{                                n
   ---> Divide o polinomio c[n] x  + ... + c[1] x + c[0] por  x - p/q  ou
        por  x + p/q                                                          }

var
  g: polinomio;
  i, k: byte;
  x: real;

begin
  x := p/q;
  g.coef[c.grau + 1] := 0;
  for k := c.grau downto 0 do
    g.coef[k] := c.coef[k] + g.coef[k + 1]*x;
  if (Abs(g.coef[0]) <= erro3) then
  begin
    sol := sol + 1;
    resp.x[sol].Re := x; resp.x[sol].Im := 0;
    resp.x[sol].metodo := metodo;
    c.grau := c.grau - 1;
    for i := 0 to c.grau do c.coef[i] := g.coef[i + 1];
  end;
  x := -p/q;
  g.coef[c.grau + 1] := 0;
  for k := c.grau downto 0 do
    g.coef[k] := c.coef[k] + g.coef[k + 1]*x;
  if (Abs(g.coef[0]) <= erro3) then
  begin
    sol := sol + 1;
    resp.x[sol].Re := x; resp.x[sol].Im := 0;
    resp.x[sol].metodo := metodo;
    c.grau := c.grau - 1;
    for i := 0 to c.grau do c.coef[i] := g.coef[i + 1];
  end;
end;

{ --------------------------------------------------------------------------- }

function MDC(m, n: longint): longint;       { Maximo divisor comum            }

var
  q, r, auxiliar: longint;

begin
  if (m < n) then
  begin
    auxiliar := m;
    m := n;
    n := auxiliar;
  end;
  repeat
    q := m div n;
    r := m mod n;
    m := n;
    n := r;
  until (r = 0);
  MDC := m;
end;

{ --------------------------------------------------------------------------- }

function MMC(m, n: longint): longint;       { Minimo multiplo comum           }

var
  x: longint;

begin
  if ((m = 0) or (n = 0)) then
    MMC := m + n
  else
  begin
    x := MDC(m, n);
    if (m*(n div x) > 2147483647) then
      MMC := 0
    else
      MMC := m*(n div x);
  end;
end;

{ --------------------------------------------------------------------------- }

procedure MMCMat(coef: PolInteiro; var teste: boolean; var d: longint);

var
  i: byte;
  DAux: longint;

begin
  d := coef.a[0];
  for i := 1 to coef.quant do
  begin
    d := MMC(coef.a[i], d);
    if (d = 0) then
    begin
      teste := false;
      Exit;
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure FormaPolar(a, b: real; var r, teta: real);

begin
  r := Sqrt(Sqr(a) + Sqr(b));
  if (a = 0) then
    if (b  > 0) then
      teta := Pi/2
    else
      if (b < 0) then teta := 3*Pi/2;
  if (b = 0) then
    if (a >= 0) then
      teta := 0
    else
      teta := Pi;
  if (a > 0) then
    if (b > 0) then
      teta := ArcTan(b/a)
    else
      if (b < 0) then teta := ArcTan(b/a) + 2*Pi;
  if ((a < 0) and (b <> 0)) then teta := ArcTan(b/a) + Pi;
end;

{ --------------------------------------------------------------------------- }

procedure VerificaSubstituicao(p: polinomio; var n: byte;
                                                       var substitui: boolean);

var
  i, j, k: byte;
  zeros: array[1..GrauMax] of byte;

begin
  substitui :=  true;
  for i := 1 to GrauMax do zeros[i] := 0;
  i := 0; k := 1;
  repeat
    while (Abs(p.coef[i]) < erro5) do
    begin
      Inc(i);
      Inc(zeros[k]);
    end;
    if zeros[k] = 0 then Inc(i);
    if (zeros[k] > 0) and (zeros[k + 1] = 0) then Inc(k);
  until i >= p.grau;
  if k = 1 then
  begin
    substitui := false;
    Exit;
  end;
  n := zeros[1] + 1;
  for i := 2 to k - 1 do n := MDC(zeros[i] + 1, n);
  for i := 0 to p.grau do
    if ((i + 1) mod n <> 1) and (Abs(p.coef[i]) > erro5) then
      substitui := false;
end;

{ --------------------------------------------------------------------------- }

procedure CalculaRaizesNEsimas(a, b: real; n: byte; var resp: solucoes;
                                                                var sol: byte);

var
  k: byte;
  r, teta, s: real;

begin
  FormaPolar(a, b, r, teta);
  s := Exp(Ln(r)/n);
  for k := 1 to n do
  begin
    resp.x[k + sol].Re := s*Cos((teta + 2*k*Pi)/n);
    resp.x[k + sol].Im := s*Sin((teta + 2*k*Pi)/n);
    resp.x[k + sol].metodo := 'Subst';
  end;
  Inc(sol, n);
end;

{ --------------------------------------------------------------------------- }

procedure CalculaCoefAuxiliar(var c: polinomio; var CoefAuxiliar: PolInteiro;
                                                           var teste: boolean);

var
  DenomAux: PolInteiro;
  i, j, k, denominador: longint;
  AchouDenAux: array[0..GrauMax] of boolean;
  aux: polinomio;
  divisao: boolean;

begin
  divisao := false;
  i := 0;
  repeat
    if ((Abs(c.coef[i]) > 2147483647) or
                              (Abs(c.coef[i] - Round(c.coef[i])) > erro4)) then
      divisao := true;
    Inc(i);
  until (i > c.grau) or divisao;
  if divisao then
  begin
    for k := 0 to c.grau do
      if (c.coef[k] <> 0) then
      begin
        i := 1; AchouDenAux[k] := false;
        repeat
          if (c.coef[k]*i > 2147483647) then
          begin
            teste := false;
            Exit;
          end;
          if (Abs(c.coef[k]*i - Round(c.coef[k]*i)) < erro4) then
          begin
            AchouDenAux[k] := true;
            DenomAux.a[k] := i;
          end;
          Inc(i);
        until AchouDenAux[k] or (i > 1000000);
        if (not AchouDenAux[k]) then
        begin
          teste := false;
          Exit;
        end;
      end;
    DenomAux.quant := c.grau;
    MMCMat(DenomAux, teste, denominador);
    if (not teste) then Exit;
    for j := 0 to c.grau do aux.coef[j] := c.coef[j]*denominador;
    for i := 0 to c.grau do
    begin
      if ((Abs(aux.coef[i]) > 2147483647) or
                          (Abs(aux.coef[i] - Round(aux.coef[i])) > erro4)) then
      begin
        teste := false;
        Exit
      end
      else
      begin
        CoefAuxiliar.a[i] := Round(aux.coef[i]);
        c.coef[i] := Round(aux.coef[i]);
      end;
    end;
  end
  else
    for j := 0 to c.grau do CoefAuxiliar.a[j] := Round(c.coef[j]);
end;

{ --------------------------------------------------------------------------- }

procedure RaizesRac(var c: polinomio; var resp:solucoes; var sol: byte;
                                              var ExecRac: boolean; lim: word);

{ ---> calcula as raizes racionais da equacao                                 }

var
  p, q: array[1..1000] of longint;
  CoefAuxiliar: PolInteiro;
  i, j, k, x, y, soma, altsoma, sp, sq: longint;
  QuantSol: byte;
  aux: polinomio;

begin
  QuantSol := sol; aux := c;
  while (c.coef[0] = 0) do
  begin
    sol := sol + 1;
    resp.x[sol].Re := 0; resp.x[sol].Im := 0;
    c.grau := c.grau - 1;
    for j := 0 to c.grau do
      c.coef[j] := c.coef[j + 1];
  end;
  CalculaCoefAuxiliar(c, CoefAuxiliar, ExecRac);
  if (not ExecRac) then Exit;
  DividePol(c, 1, 1, resp, sol, 'Racnl');
  soma := 0; altsoma := 0;
  for i := 0 to c.grau do
  begin
    soma := soma + CoefAuxiliar.a[i];
    if Odd(i) then
      altsoma := altsoma - CoefAuxiliar.a[i]
    else
      altsoma := altsoma + CoefAuxiliar.a[i];
  end;
  x := Abs(CoefAuxiliar.a[0]); y := Abs(CoefAuxiliar.a[c.grau]);
  sp := 1; sq := 1;
  p[1] := x; q[1] := y;
  for i := 1 to Min(lim, x div 2) do
    if (x mod i = 0) then
    begin
      sp := sp + 1;
      p[sp] := i;
    end;
  for j := 1 to Min(lim, y div 2) do
    if (y mod j = 0) then
    begin
      sq := sq + 1;
      q[sq] := j;
    end;
  for i := 1 to sp do
    for j := 1 to sq do
      if ((p[i] <> q[j]) and
          ((soma mod (p[i] - q[j]) = 0) or (soma mod (p[i] + q[j]) = 0)) and
          ((altsoma mod (p[i] + q[j]) = 0) or
          (altsoma mod (p[i] - q[j]) = 0))) then
        DividePol(c, p[i], q[j], resp, sol, 'Racnl');
  if (QuantSol = sol) then ExecRac := false;
end;

{ --------------------------------------------------------------------------- }

procedure Gauss(var c: polinomio; var resp: solucoes; var sol: byte;
                                            var ExecGauss: boolean; lim: word);

{ ---> procura solucoes do tipo a + bi onde a, b sao inteiros                 }

var
  CoefAuxiliar: PolInteiro;
  SomaCoef, i, j, k, soma, quoc: longint;
  QuantSol: byte;
  DivExata: boolean;
  f1: real;

begin
  DivExata := false; f1 := 0;
  CalculaCoefAuxiliar(c, CoefAuxiliar, ExecGauss);
  if (not ExecGauss) then Exit;
  for i := 0 to c.grau do
    f1 := f1 + CoefAuxiliar.a[i];
  if (Abs(f1) > 2147483647) then
  begin
    ExecGauss := false;
    Exit;
  end;
  SomaCoef := Round(f1);
  QuantSol := sol;
  for j := 1 to Min(lim, Round(Sqrt(Abs(CoefAuxiliar.a[0])))) do
    if ((CoefAuxiliar.a[0] mod j*j = 0) and (SomaCoef mod (1 + j*j) = 0)) then
    begin
      DivPolinomio(c, 0, j*j, DivExata, resp, sol);
      if DivExata then
      begin
        resp.x[sol - 1].metodo := 'Gauss';
        resp.x[sol].metodo := 'Gauss';
        if (c.grau = 0) then Exit;
        for i := 0 to c.grau do
          CoefAuxiliar.a[i] := Round(c.coef[i]);
        SomaCoef := SomaCoef div (1 + j*j);
      end;
    end;
  for i := 1 to Min(lim, Trunc(Sqrt(Abs(CoefAuxiliar.a[0])/2)) + 1) do
    for j := 1 to Min(lim, Trunc(Sqrt(Abs(CoefAuxiliar.a[0])))) do
    begin
      soma := i*i + j*j;
      if (CoefAuxiliar.a[0] mod (i*i + j*j) = 0) then
      begin
        quoc := CoefAuxiliar.a[0] div soma;
        if (((CoefAuxiliar.a[1] + 2*i*quoc) mod soma = 0) and
                                      (SomaCoef mod (1 + soma - 2*i) = 0)) then
          DivPolinomio(c, -2*i, soma, DivExata, resp, sol);
        if DivExata then
        begin
          resp.x[sol - 1].metodo := 'Gauss';
          resp.x[sol].metodo := 'Gauss';
          if (c.grau = 0) then Exit;
          for k := 0 to c.grau do
            CoefAuxiliar.a[k] := Round(c.coef[k]);
          SomaCoef := SomaCoef div (1 + soma - 2*i);
          DivExata := false;
        end;
      end;
      if (((CoefAuxiliar.a[1] - 2*i*quoc) mod soma = 0) and
                                      (SomaCoef mod (1 + soma + 2*i) = 0)) then
        DivPolinomio(c, 2*i, soma, DivExata, resp, sol);
      if DivExata then
      begin
        resp.x[sol - 1].metodo := 'Gauss';
        resp.x[sol].metodo := 'Gauss';
        if (c.grau = 0) then Exit;
        for k := 0 to c.grau do
          CoefAuxiliar.a[k] := Round(c.coef[k]);
        SomaCoef := SomaCoef div (1 + soma + 2*i);
        DivExata := false;
      end;
      if (j > Trunc(Sqrt(abs(CoefAuxiliar.a[0]))) div 2) then
      begin
        if (((CoefAuxiliar.a[1] + 2*j*quoc) mod soma = 0) and
                                      (SomaCoef mod (1 + soma - 2*j) = 0)) then
          DivPolinomio(c, -2*j, soma, DivExata, resp, sol);
        if DivExata then
        begin
          resp.x[sol - 1].metodo := 'Gauss';
          resp.x[sol].metodo := 'Gauss';
          if (c.grau = 0) then Exit;
          for k := 0 to c.grau do
            CoefAuxiliar.a[k] := Round(c.coef[k]);
          SomaCoef := SomaCoef div (1 + soma - 2*j);
          DivExata := false;
        end;
        if (((CoefAuxiliar.a[1] - 2*j*quoc) mod soma = 0) and
                                     (SomaCoef mod (1 + soma + 2*j) = 0)) then
          DivPolinomio(c, 2*j, soma, DivExata, resp, sol);
        if DivExata then
        begin
          resp.x[sol - 1].metodo := 'Gauss';
          resp.x[sol].metodo := 'Gauss';
          if (c.grau = 0) then Exit;
          for k := 0 to c.grau do
            CoefAuxiliar.a[k] := Round(c.coef[k]);
          SomaCoef :=  SomaCoef div (1 + soma + 2*j);
          DivExata := false;
        end;
      end;
    end;
  if (QuantSol = sol) then ExecGauss := false;
end;

{ --------------------------------------------------------------------------- }

procedure Bairstow(var a: polinomio; var resp: solucoes; var sol: byte;
                   var resolveu: boolean; lim: byte);

{                                                     n
  ---> determina reais  p[w] e  q[w] tais que    a[n]x  + ... + a[1]x + a[0]
                      2
  seja divisivel por x  - p[w] x - q[w] . E' utilizado o teorema da  Analise
  Numerica  conhecido como  "Algoritmo  de  Bairstow".                        }

procedure Reinicia(var w: integer; var aprox: byte; var p0, q0: real;
                   var resolveu: boolean; lim: byte);

begin { ---> Reinicia }
  w := -1;
  aprox := aprox + 1;
  if aprox > lim then
  begin
    resolveu := false;
    Exit
  end;
  p0 := Trunc((aprox + 3)/4);
  q0 := Trunc((aprox + 3)/4);
  case (aprox mod 4) of
    0 : p0 := -p0;
    2 : begin
          p0 := -p0;
          q0 := -q0;
        end;
    3 : q0 := -q0;
  end;
end; { ---> Reinicia }

type
  linha = array[-2..101] of real;

var
  disc: real;
  b, c, p, q: linha;
  aprox: byte;
  i, j, k, w: integer;
  DivExata: boolean;

begin { ---> Bairstow }
  DivExata := false;
  aprox := 0;
  p[0] := 0; q[0] := 0;
  p[-1] := 0; q[-1] := 0;
  b[-1] := 0; b[-2] := 0;
  c[-1] := 0; c[-2] := 0;
  w := -1;
  repeat
    w := w + 1;
    for j := 0 to a.grau do
      b[j] := a.coef[a.grau - j] + p[w]*b[j - 1] + q[w]*b[j - 2];
    for j := 0 to a.grau - 1 do
      c[j] := b[j] + p[w]*c[j - 1] + q[w]*c[j - 2];
    disc := Sqr(c[a.grau - 2]) - c[a.grau - 1]*c[a.grau - 3];
    if ((Abs(disc) <= erro6) or (w > 100)) then
    begin
      Reinicia(w, aprox, p[0], q[0], resolveu, lim);
      if not resolveu then Exit;
    end
    else
    begin
      p[w + 1] := p[w] - (b[a.grau - 1]*c[a.grau - 2] -
                                                 c[a.grau - 3]*b[a.grau])/disc;
      q[w + 1] := q[w] - (c[a.grau - 2]*b[a.grau] -
                                             b[a.grau - 1]*c[a.grau - 1])/disc;
    end;
    if ((Abs(p[w+1] - p[w]) <= erro5) and (Abs(q[w+1] - q[w]) <= erro5)) then
    begin
      DivPolinomio(a, -p[w], -q[w], DivExata, resp, sol);
      if not DivExata then
      begin
        Reinicia(w, aprox, p[0], q[0], resolveu, lim);
        if not resolveu then Exit;
      end
      else
      begin
        resp.x[sol - 1].metodo := 'Brstw';
        resp.x[sol].metodo := 'Brstw';
      end;
    end;
  until DivExata;
end; { ---> Bairstow }

{ --------------------------------------------------------------------------- }

procedure ResolveEquacao(a: polinomio; op: byte; var resp: solucoes);

var
  substitui, ExecGauss, ExecRac, inteiro: boolean;
  sol, n, i, aux, LimInt, LimGauss, LimBairstow: byte;
  SomaCoef: longint;
  s: real;
  LimRac: word;
  RespAux: solucoes;
  EqAux: polinomio;

begin
  if (op = 0) then
  begin
    LimInt := 50;
    LimBairstow := 10;
    LimRac := 1000;
    LimGauss := 50;
  end
  else
  begin
    LimInt := 100;
    LimBairstow := 80;
    LimRac := 10000;
    LimGauss := 200;
  end;
  sol := 0;
  resp.resolveu := true;
  inteiro := true;
  for i := 0 to a.grau do
    if (a.coef[i] <> Int(a.coef[i])) then inteiro := false;
  if (Abs(a.coef[0]) > 2147483647) then inteiro := false;
  if inteiro then
  begin
    s := 0;
    for i := 0 to a.grau do s := s + a.coef[i];
    if (Abs(s) > 2147483647) then inteiro := false else SomaCoef := Trunc(s);
  end;
  for i := 0 to LimInt do
  begin
    repeat
      aux := sol;
      if ((not inteiro) or (i <= 1) or ((Trunc(a.coef[0]) mod i = 0) and
              ((SomaCoef mod (i - 1) = 0) or (SomaCoef mod (i + 1) = 0)))) then
        DividePol(a, i, 1, resp, sol, 'Intei');
      if (a.grau = 0) then Exit;
    until (aux = sol);
  end;
  ExecRac := true;
  ExecGauss := true;
  repeat
    if ExecRac then
      RaizesRac(a, resp, sol, ExecRac, LimRac)
    else
      if ExecGauss then
        Gauss(a, resp, sol, ExecGauss, LimGauss)
      else
      begin
        VerificaSubstituicao(a, n, substitui);
        if substitui then
        begin
          EqAux.grau := a.grau div n;
          for i := 0 to EqAux.grau do EqAux.coef[i] :=  a.coef[i*n];
          ResolveEquacao(EqAux, op, RespAux);
          if RespAux.resolveu then
            for i := 1 to EqAux.grau do
              CalculaRaizesNEsimas(RespAux.x[i].Re, RespAux.x[i].Im, n,
                                                                    resp, sol);
          resp.resolveu := (sol >= a.grau);
          if resp.resolveu then a.grau := 0;
        end;
        if (not substitui) or (not resp.resolveu) then
          Bairstow(a, resp, sol, resp.resolveu, LimBairstow);
      end;
    if (not resp.resolveu) then Exit;
  until (a.grau <= 2);

  if a.grau = 2 then
  begin
    SegGrau(a.coef[1]/a.coef[2], a.coef[0]/a.coef[2], resp, sol);
    resp.x[sol - 1].metodo := 'SegGr';
    resp.x[sol].metodo := 'SegGr';
  end
  else
    if a.grau = 1 then
    begin
      sol := sol + 1;
      resp.x[sol].Re := -a.coef[0]/a.coef[1];
      resp.x[sol].Im := 0;
      resp.x[sol].metodo := 'PriGr';
    end;
end;

{ --------------------------------------------------------------------------- }

var
  arq: file of real;

begin
  Assign(arq, 'EQUACAO.ERR');
  {$I-}
  Reset(arq);
  {$I+}
  if (IOResult = 0) then
  begin
    Read(arq, erro1, erro2, erro3, erro4, erro5, erro6);
    Close(arq);
  end;
end.

{ =========================================================================== }


