{ =========================================================================== }
{                                 MATRIZES.PAS                                }
{ =========================================================================== }

unit Matrizes;

{ ---> procedimentos e funcoes usados para fazer operacoes com matrizes tais
       como calculo da matriz inversa, determinante, escalonamento, nulidade
       entrada e saida de matrizes, resolucao de sistemas lineares

       (C) 1989, 1990 Lenimar Nunes de Andrade
           CCENDM03@BRUFPB.BITNET                                             }

{$E+,N+,F+,R+,V-,B-,S+}

interface

uses
  FPTCBas;

procedure FormataMatriz(mat: matriz; var tam1, tam2: byte);
procedure ReformataMatriz(mat: matriz; var tam1, tam2: byte);
procedure GeraMatrizAleatoria(var mat: matriz; op: byte);
procedure MostraMatriz(mat: matriz; op: InfoPress);
procedure LeMatriz(var mat: matriz; op: byte);
procedure EscalonaMatriz(var mat: matriz; opcao: byte);
procedure InverteMatriz(var mat: matriz; var inversivel: boolean);
procedure MultiplicaMatrizes(mat1, mat2: matriz; var mat3: matriz);
procedure ResolveSistema(mat: matriz; var solucao: resposta);
procedure ElevaMatriz(var mat: matriz; n: shortint);
procedure CalculaConjugada(var mat: matriz; p: matriz; var invers: boolean);
function Det(mat: matriz): real;
function Nulidade(mat: matriz; lambda: real; n: byte): byte;
function Nul(mat: matriz): byte;
procedure LeMatrizDisco(var mat: matriz; nome: string; var HErro: boolean);
procedure SalvaMatriz(mat: matriz; op: byte; nome: string; tela: TelaPtr);
procedure Moldura(mat: matriz; imp: InfoPress; caso: byte; tela: TelaPtr);
procedure LeTeclado(var mat: matriz);
function LI(mat: matriz): boolean;
procedure MatrizTransposta(var mat: matriz);
procedure MatrizMudancaDeBase(mat1, mat2: matriz; var mat3: matriz;
                                                        var SaoBases: boolean);
procedure MatrizRelBase(mat, b: matriz; var NMat: matriz);

{ =========================================================================== }

implementation

uses
  Crt, Printer;

type
  InfoVarLivre = record
    VariavelLivre: array[1..MaxN] of byte;
    QuantVarLivres: byte;
    IndVarNaoLivre: set of byte;
  end;

  ind = array[1..100] of string[14];

const
  erro1: real = 1E-10;   { usado no calculo da matriz inversa }
  erro2: real = 1E-10;   { usado no calculo do determinante   }
  erro3: real = 1E-10;   { usado no calculo da nulidade       }
  decimais: byte = 4;    { quantidade de casas decimais       }

{ --------------------------------------------------------------------------- }

procedure FormataMatriz(mat: matriz; var tam1, tam2: byte);

{ ---> determina um formato tam1:tam2 para os elementos da matriz mat         }

var
  i, j: byte;
  inteiro: boolean;

begin
  inteiro := true;
  tam1 := 1;
  for i := 1 to mat.m do
    for j := 1 to mat.n do
    begin
      if (Abs(mat.a[i, j]) > 2147483647) then
      begin
        tam1 := 0;
        tam2 := decimais;
        Exit;
      end
      else
      begin
        if (mat.a[i, j] <> Int(mat.a[i, j])) then inteiro := false;
        mat.a[i, j] := Int(mat.a[i, j]);
        if mat.a[i, j] <> 0 then
          if (Ln(Abs(mat.a[i, j]))/Ln(10) > tam1) then
            tam1 := Round(Ln(Abs(mat.a[i, j]))/Ln(10));
      end;
    end;
    if inteiro then tam2 := 0 else tam2 := decimais;
    tam1 := tam1 + tam2 + 4;
end;

{ --------------------------------------------------------------------------- }

procedure ReformataMatriz(mat: matriz; var tam1, tam2: byte);

{ ---> diminui, se possivel, o valor de tam1 calculado por FormataMatriz      }

var
  min, i, j, k: byte;
  x: string;

begin
  min := 100;
  for i := 1 to mat.m do
    for j := 1 to mat.n do
    begin
      if (Abs(mat.a[i, j]) > 2147483647) then Exit;
      x := Rac(mat.a[i, j], tam1, tam2);
      k := 0;
      while (x[k + 1] = ' ') do Inc(k);
      if (k = 1) then Exit;
      if (k < min) then min := k;
    end;
    tam1 := tam1 - min + 2;
end;

{ --------------------------------------------------------------------------- }

procedure GeraMatrizAleatoria(var mat: matriz; op: byte);

var
  i, j: byte;
  sinal: shortint;

const
  L: byte = 5;

begin
  if (op = 1) then
  begin
    Write('Quantidade de linhas : '); Readln(mat.m);
    Write('Quantidade de colunas : '); Readln(mat.n);
  end
  else
    if (op = 2) then
    begin
      Write('Ordem da matriz : '); Readln(mat.m); mat.n := mat.m;
    end;
  if (mat.m >= 10) then
    L := 3
  else
    if (mat.m <= 5) then L := 9;
  Randomize;
  for i := 1 to mat.m do
    for j := 1 to mat.n do
    begin
      sinal := Random(L);
      if Odd(sinal) then sinal := -1 else sinal := 1;
      mat.a[i, j] := sinal*Random(L);
    end;
end;

{ --------------------------------------------------------------------------- }

procedure MostraMatriz(mat: matriz; op: InfoPress);

{ ---> mostra a matriz mat no formato real ou racional, na tela ou no papel   }

procedure MostraMatrizColuna(mat: matriz; op: InfoPress; tam1, tam2: byte);

var
  i, j, ordem: byte;
  ch: char;

begin { ---> MostraMatrizColuna }
  ordem := mat.m; if (ordem > 20) then Exit;
  Window(1, 1, 25, ordem + 5);
  Cor(7, 0);
  j := 1;
  repeat
    GoToXY(1, 1);
    Write(#218, #196, ' Coluna ', j:2, ' da matriz ', #196, #191);
    for i := 1 to (ordem + 2) do
    begin
      GoToXY(1, i + 1); Write(#179, Spc(23), #179);
    end;
    Write(#192, Repete(3, #196), ' Ordem = ', ordem:2, ' x ', ordem:2, ' ');
    Write(Repete(3, #196), #217);
    Window(11, 22, 69, 22);
    if (mat.m <= 17) then
      Write(MatMenu[14])
    else
    begin
      Window(28, 22, 69, 22);
      Write(MatMenu[15]);
    end;
    Window(1, 1, 25, ordem + 5);
    for i := 1 to mat.n do
    begin
      GoToXY(10, i + 2);
      if (op.racional = 0) then
        if tam1 > 0 then Write(mat.a[i, j]:tam1:tam2) else Write(mat.a[i, j]:0)
      else
        Write(Rac(mat.a[i, j], tam1, tam2));
    end;
    Window(11, 22, 69, 22);
    if (mat.m <= 17) then
      Write(MatMenu[10])
    else
    begin
      Window(28, 22, 69, 22);
      Write(MatMenu[11]);
    end;
    Window(1, 1, 25, ordem + 5);
    repeat
      ch := ReadKey;
    until (ch = ESC) or (ch = #0);
    if (ch = #0) then
    begin
      ch := ReadKey;
      if (ch = SetaPEsquerda) then
      begin
        if (j > 1) then Dec(j) else Write(#7)
      end
    else
      if (ch = SetaPDireita) then
         if (j < ordem) then Inc(j) else Write(#7);
    end;
  until (ch = ESC);
end; { ---> MostraMatrizColuna }

var
  TipoImp, horiz, i, j, tam1, tam2: byte;
  espacos: integer;

begin { ---> MostraMatriz  }
  FormataMatriz(mat, tam1, tam2);
  if (op.racional <> 0) then ReformataMatriz(mat, tam1, tam2);
  if (mat.m = mat.n) and (mat.m <= 5) then
    TipoImp := 1
  else
    TipoImp := 0;
  if (op.tela = 1) and (TipoImp = 0) and (op.opcao = 1) then
  begin
    if (tam1 > 10) then tam1 := 0;
    MostraMatrizColuna(mat, op, tam1, tam2);
    Exit;
  end;
  if (op.largura > tam1*mat.n + 2) then
    espacos := Trunc((op.largura - tam1*mat.n)/2)
  else
  begin
    tam1 := 0;
    espacos := 1;
  end;
  if (tam1 = 0) and (op.tela = 1) then
  begin
    TipoImp := 0;
    if mat.m <= 5 then Writeln;
    if mat.m <= 3 then Writeln;
    case mat.m of
      2 : espacos := 18;
      3 : espacos := 13;
      4 : espacos := 7;
      5 : espacos := 2;
    end;
  end;
  if (op.tela = 1) and (TipoImp = 1) and (op.opcao = 1) then
  begin
    if (mat.m >= 4) then horiz := 1 else horiz := 2;
    GoToXY(espacos, horiz);
    Writeln(#218, ' ':(mat.m*tam1 + 2), #191);
  end;
  for i := 1 to mat.m do
  begin
    if (op.tela = 1) then
      if (TipoImp <> 1) or (op.opcao <> 1) then
        Write(' ':espacos)
      else
      begin
        Write(' ':(espacos - 1));
        Write(#179);
      end
    else
      Write(Lst, ' ':espacos);
    for j := 1 to mat.n do
      if (op.racional = 0) then
        if (tam1 = 0) then
          if (op.tela = 1) then
            Write(mat.a[i, j]:0, ' ')
          else
            Write(Lst, mat.a[i, j]:0, ' ')
        else
          if (op.tela = 1) then
            Write(mat.a[i, j]:tam1:tam2)
          else
            Write(Lst, mat.a[i, j]:tam1:tam2)
      else
        if (op.tela = 1) then
          Write(Rac(mat.a[i, j], tam1, tam2))
        else
          Write(Lst, Rac(mat.a[i, j], tam1, tam2));
    if (op.tela = 1) then
      if (TipoImp <> 1) or (op.opcao <> 1) then
        Writeln
      else
        Writeln('  ', #179)
    else
      Writeln(Lst);
  end;
  if (op.tela = 1) and (TipoImp = 1) and (op.opcao = 1) then
    Write(' ':(espacos - 1), #192, ' ':(mat.m*tam1 + 2), #217);
end; { ---> MostraMatriz }

{ --------------------------------------------------------------------------- }

procedure LeMatriz(var mat: matriz; op: byte);

{ ---> le os elementos da matriz mat (que podem ser expressoes aritmeticas co-
       mo -5/3, 2^10, etc.)                                                   }

var
  x: array[1..MaxN] of string[20];
  y: string[80];
  i, j, k, simb1, simb2, compr: byte;
  cod: integer;
  valor1, valor2: real;
  sinal: shortint;

begin
  if (op = 1) then
  begin
    Write('Quantidade de linhas : '); Readln(mat.m);
    Write('Quantidade de colunas : '); Readln(mat.n);
  end
  else
    if (op = 2) then
    begin
      Write('Ordem da matriz : '); Readln(mat.m); mat.n := mat.m;
    end;
  if (op <> 3) then
  begin
    Writeln;
    Writeln('Forneca os elementos da matriz : ');
    Writeln;
  end;
  for i := 1 to mat.m do
  begin
    if (op <> 3) then
      Write('Linha ', i, ' ', #16 ,' ')
    else
      Write('[', i:2, '] ', #16, ' ');
    Readln(y);
    k := 1;
    while(y[k] = ' ') do Inc(k);
    if (k >= 2) then y := Copy(y, k, Length(y));
    k := 1;
    for j := 1 to mat.n do
    begin
      x[j] := '';
      while (y[k] <> ' ') and  (k <= Length(y)) do
      begin
        if (y[k] = ',') then y[k] := '.';
        x[j] := x[j] + y[k];
        Inc(k);
      end;
      while (y[k] = ' ') do Inc(k);
      compr := Length(x[j]);
      Val(x[j], mat.a[i, j], cod);
      simb1 := Pos('/', x[j]);
      simb2 := Pos('^', x[j]);
      if (simb1 > 0) then
      begin
        Val(Copy(x[j], 1, simb1 - 1), valor1, cod);
        Val(Copy(x[j], simb1 + 1, compr - simb1), valor2, cod);
        if (valor2 = 0) then Halt;
        mat.a[i, j] := valor1/valor2;
      end
      else
        if (simb2 > 0) then
        begin
          if (x[j][1] = '-') then
          begin
            sinal := -1;
            x[j] := Copy(x[j], 2, Length(x[j]));
            simb2 := Pos('^', x[j]);
          end
          else
            sinal := 1;
          Val(Copy(x[j], 1, simb2 - 1), valor1, cod);
          Val(Copy(x[j], simb2 + 1, compr - simb2), valor2, cod);
          if (valor1 <= 0) then Halt;
          mat.a[i, j] := sinal*Exp(valor2*Ln(valor1));
        end;
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure EscalonaMatriz(var mat: matriz; opcao: byte);

{ ---> escalona a matriz mat

       opcao 1: nao e' feita divisao no escalonamento; matriz de elementos
                inteiros continuara' com elementos inteiros apos o escalo-
                namento.
       opcao 2: e' feita divisao e em cada coluna e' escolhido o maior ele-
                mento em modulo (metodo da eliminacao de Gauss com pivote-
                amento parcial).
       opcao 3: e' feita divisao mas nao e' escolhido o elemento de  maior
                modulo em cada coluna (metodo da eliminacao de Gauss).        }

var
  i, j, k, linha, col, NumZeros, MaiorElemento: byte;
  elem1, elem2, aux: real;

begin
  for i := 1 to mat.m do
    for j := 1 to mat.n do
      Aproxima(mat.a[i, j]);
  linha := 1; col := 1;
  while ((linha <= mat.m) and (col <= mat.n)) do
  begin
    MaiorElemento := linha;
    NumZeros := 0;
    for i := mat.m downto linha do
      if (mat.a[i, col] <> 0) then
      begin
        if (opcao <> 2) then
          MaiorElemento := i
        else
          if (Abs(mat.a[i, col]) > Abs(mat.a[MaiorElemento, col])) then
            MaiorElemento := i
      end
      else
        NumZeros := NumZeros + 1;
    if (MaiorElemento <> linha) then
      for j := col to mat.n do
      begin
        aux := mat.a[linha, j];
        mat.a[linha, j] := mat.a[MaiorElemento, j];
        mat.a[MaiorElemento, j] := aux;
      end;
    if (NumZeros < mat.m - linha) then
    begin
      for j := linha + 1 to mat.m do     { escalonamento da matriz }
      begin
        elem1 := mat.a[j, col]; elem2 := mat.a[linha, col];
        for k := col to mat.n do
          if (opcao = 1) then
          begin
            mat.a[j, k] := elem2*mat.a[j, k] - elem1*mat.a[linha, k];
            Aproxima(mat.a[j, k]);
          end
          else
          begin
            mat.a[j, k] := mat.a[j, k] - elem1/elem2*mat.a[linha, k];
            Aproxima(mat.a[j, k]);
          end;
      end;
    end;
    if (NumZeros <= mat.m - linha) then linha := linha + 1;
    col := col + 1;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure InverteMatriz(var mat: matriz; var inversivel: boolean);

{ ---> calcula a inversa da matriz mat                                        }

var
  elem, col, i, j, k, ordem: byte;
  aux: matriz;
  h: real;

begin
  if (mat.m <> mat.n) then Exit;
  inversivel := true; ordem := mat.m;
  aux.m := ordem; aux.n := ordem;
  for i := 1 to ordem do
    for j := 1 to ordem do
    begin
      Aproxima(mat.a[i, j]);
      if (i = j) then aux.a[i, j] := 1 else aux.a[i, j] := 0;
    end;
  for col := 1 to ordem - 1 do
  begin
    elem := col;
    for i := ordem downto col do
      if (mat.a[i, col] <> 0) then elem := i;
    if (mat.a[elem, col] <> 0) then
    begin
      if (elem <> col) then
        for j := 1 to ordem do
        begin
          h := mat.a[elem, j];
          mat.a[elem, j] := mat.a[col, j];
          mat.a[col, j] := h;
          h := aux.a[elem, j];
          aux.a[elem, j] := aux.a[col, j];
          aux.a[col, j] := h;
        end;
      for j := (col + 1) to ordem do
      begin
        h := mat.a[j, col]/mat.a[col, col];
        for k := 1 to ordem do
        begin
          mat.a[j, k] := mat.a[j, k] - h*mat.a[col, k];
          aux.a[j, k] := aux.a[j, k] - h*aux.a[col, k];
          Aproxima(mat.a[j, k]);
          Aproxima(aux.a[j, k]);
        end;
      end;
    end;
  end;
  for k := 1 to ordem do
  begin
    h := mat.a[k, k];
    if (Abs(h) <= erro1) then
    begin
      inversivel := false;
      Exit;
    end;
    for j := 1 to ordem do
    begin
      mat.a[k, j] := mat.a[k, j]/h;
      aux.a[k, j] := aux.a[k, j]/h;
    end;
  end;
  for col := 2 to ordem do
    for i := 1 to (col - 1) do
    begin
      h := mat.a[i, col];
      Aproxima(h);
      for j := 1 to ordem do
      begin
        mat.a[i, j] := mat.a[i, j] - h*mat.a[col, j];
        aux.a[i, j] := aux.a[i, j] - h*aux.a[col, j];
        Aproxima(aux.a[i, j]);
      end;
    end;
  mat := aux;
end;

{ --------------------------------------------------------------------------- }

procedure MultiplicaMatrizes(mat1, mat2: matriz; var mat3: matriz);

{ ---> multiplica as matrizes mat1 e mat2 e devolve mat3 como resultado       }

var
  i, j, k: byte;

begin
  if (mat1.n <> mat2.m) then Exit;
  mat3.m := mat1.m; mat3.n := mat2.n;
  for i := 1 to mat1.m do
    for j := 1 to mat2.n do
    begin
      mat3.a[i, j] := 0;
      for k := 1 to mat1.m do
        mat3.a[i, j] := mat3.a[i, j] + mat1.a[i, k]*mat2.a[k, j];
    end;
end;

{ --------------------------------------------------------------------------- }

procedure ResolveSistemaDeterminado(mat: matriz; var sol: vetor);

{ ---> resolve o sistema linear determinado cuja matriz escalonada e' mat     }

var
  i, k: byte;
  soma: real;

begin
  for i := (mat.n - 1) downto 1 do
  begin
    soma := mat.a[i, mat.n];
    for k := (mat.n - 1) downto (i + 1) do
      soma := soma - mat.a[i, k]*sol[k];
    sol[i] := 1/mat.a[i, i]*soma;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure CalculaVariaveisLivres(mat: matriz; var VLivre: InfoVarLivre);

{ ---> determina as variaveis livres do sistema linear cuja matriz escalonada
       e' mat                                                                 }

var
  i, elem: byte;

begin
  VLivre.IndVarNaoLivre := [];
  i := 1;
  while (i <= mat.m) do
  begin
    elem := i - 1;
    repeat
      Inc(elem);
    until (mat.a[i, elem] <> 0) or (elem = mat.n);
    if (elem < mat.n) then
      VLivre.IndVarNaoLivre := VLivre.IndVarNaoLivre + [elem];
    if (elem >= mat.n - 1) then i := mat.m;
    Inc(i);
  end;
  VLivre.QuantVarLivres := 0;
  for i := 1 to mat.n - 1 do
    if not (i in VLivre.IndVarNaoLivre) then
    begin
      Inc(VLivre.QuantVarLivres);
      VLivre.VariavelLivre[VLivre.QuantVarLivres] := i;
    end;
end;

{ --------------------------------------------------------------------------- }

procedure ResolveSistemaHomogeneo(mat: matriz; VLivre: InfoVarLivre;
                                                        var solucao: resposta);

{ ---> resolve o sistema linear homogeneo definido pela matriz escalonada mat.
       E' encontrada a solucao geral.                                         }

var
  i, j, k: byte;
  aux: matriz;
  sol: vetor;

begin
  solucao.x.m := VLivre.QuantVarLivres;
  solucao.x.n := mat.n - 1;
  aux.m := mat.m - VLivre.QuantVarLivres;
  aux.n := mat.n - VLivre.QuantVarLivres;
  j := 1;
  while (j < aux.n) do
  begin
    for k := 1 to mat.n - 1 do
      if (k in VLivre.IndVarNaoLivre) then
      begin
        for i := 1 to aux.m do aux.a[i, j] := mat.a[i, k];
        j := j + 1;
      end;
  end;
  for k := 1 to VLivre.QuantVarLivres do
  begin
    for j := 1 to aux.m do
      aux.a[j, aux.n] := -mat.a[j, VLivre.VariavelLivre[k]];
      ResolveSistemaDeterminado(aux, sol);
      j := 1;
      for i := 1 to mat.n - 1 do
      begin
        if (i in VLivre.IndVarNaoLivre) then
        begin
          solucao.x.a[k, i] := sol[j];
          j := j + 1;
        end
        else
          if i = VLivre.VariavelLivre[k] then
            solucao.x.a[k, i] := 1
          else
            solucao.x.a[k, i] := 0;
      end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure EncontraSolucaoParticular(mat: matriz; VLivre: InfoVarLivre;
                                                        var solucao: resposta);

{ ---> determina uma solucao particular do sistema linear definido pela matriz
       escalonada mat                                                         }

var
  linha, elem, i, j, k: byte;
  aux: matriz;
  sol: vetor;

begin
  if (VLivre.QuantVarLivres = 0) then
  begin
    ResolveSistemaDeterminado(mat, solucao.SolParticular);
    Exit;
  end;
  aux.m := mat.m - VLivre.QuantVarLivres;
  aux.n := mat.n - VLivre.QuantVarLivres;
  j := 1;
  while (j < aux.n) do
  begin
    for k := 1 to mat.n - 1 do
      if (k in VLivre.IndVarNaoLivre) then
      begin
        for i := 1 to aux.m do aux.a[i, j] := mat.a[i, k];
        j := j + 1;
      end;
  end;
  for j := 1 to aux.m do
    aux.a[j, aux.n] := mat.a[j, mat.n];
  ResolveSistemaDeterminado(aux, sol);
  j := 1;
  for i := 1 to mat.n - 1 do
  begin
    if (i in VLivre.IndVarNaoLivre) then
    begin
      solucao.SolParticular[i] := sol[j];
      j := j + 1;
    end
    else
      solucao.SolParticular[i] := 0;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure ResolveSistema(mat: matriz; var solucao: resposta);

{ ---> resolve o sistema linear definido pela matriz mat                      }

var
  aux: matriz;
  i, j, elem: byte;
  saux: resposta;
  VLivre: InfoVarLivre;

begin
  EscalonaMatriz(mat, 2);
  solucao.impossivel := false;
  for i := 1 to mat.m do
  begin
    elem := 0;
    repeat
      Inc(elem);
    until ((elem = mat.n + 1) or (mat.a[i, elem] <> 0));
    if (elem = mat.n) then
    begin
      solucao.impossivel := true;
      Exit;
    end;
  end;
  if (mat.m > mat.n - 1) then
  begin
    solucao.EqDemais := true;
    mat.m := mat.n - 1;
  end
  else
    solucao.EqDemais := false;
  CalculaVariaveisLivres(mat, VLivre);
  solucao.x.m := VLivre.QuantVarLivres;
  solucao.x.n := mat.n - 1;
  if (mat.m < mat.n - 1) then
  begin
    for i := (mat.m + 1) to (mat.n - 1) do
      for j := 1 to mat.n do mat.a[i, j] := 0;
    mat.m := mat.n - 1;
  end;
  if (VLivre.QuantVarLivres = 0) then
    solucao.SistDet := true
  else
    solucao.SistDet := false;
  solucao.homogeneo := true;
  for i := 1 to mat.m do
    if (mat.a[i, mat.n] <> 0) then solucao.homogeneo := false;
  if solucao.homogeneo then
    for i := 1 to mat.n - 1 do solucao.SolParticular[i] := 0
  else
    EncontraSolucaoParticular(mat, VLivre, solucao);
  if (not solucao.SistDet) then
  begin
    aux := mat;
    for i := 1 to mat.m do aux.a[i, mat.n] := 0;
    ResolveSistemaHomogeneo(aux, VLivre, saux);
    solucao.x := saux.x
  end;
end;

{ --------------------------------------------------------------------------- }

procedure ElevaMatriz(var mat: matriz; n: shortint);

var
  i, j: byte;
  aux: matriz;
  inversivel: boolean;

begin
  if (mat.m <> mat.n) then Exit;
  aux.m := mat.m; aux.n := mat.n;
  for i := 1 to aux.m do
    for j := 1 to aux.n do
      if (i = j) then aux.a[i, j] := 1 else aux.a[i, j] := 0;
  if (n < 0) then
  begin
    InverteMatriz(mat, inversivel);
    if not inversivel then Exit;
  end;
  for i := 1 to Abs(n) do
    MultiplicaMatrizes(aux, mat, aux);
  mat := aux;
end;

{ --------------------------------------------------------------------------- }

procedure CalculaConjugada(var mat: matriz; p: matriz; var invers: boolean);

var
  inversivel: boolean;
  aux: matriz;

begin
  aux := p;
  InverteMatriz(aux, invers);
  if (not invers) then Exit;
  MultiplicaMatrizes(aux, mat, aux);
  MultiplicaMatrizes(aux, p, mat);
end;

{ --------------------------------------------------------------------------- }

function Det(mat: matriz): real;

{ ---> calcula o determinante da matriz mat                                   }

var
  inversao, ordem, elem, i, j, k: byte;
  d, aux: real;

begin
  if (mat.m <> mat.n) then Exit;
  inversao := 0;
  ordem := mat.m;
  for i := 1 to (ordem - 1) do
  begin
    elem := i;
    for j := ordem downto i do
      if (mat.a[j, i] <> 0) then elem := j;
    if (Abs(mat.a[elem, i]) <= erro2) then
    begin
      Det := 0;
      Exit;
    end
    else
    begin
      if (elem <> i) then
      begin
        Inc(inversao);
        for j := i to ordem do
        begin
          aux := mat.a[elem, j];
          mat.a[elem, j] := mat.a[i, j];
          mat.a[i, j] := aux;
        end;
      end;
      for j := (i + 1) to ordem do
      begin
        aux := mat.a[j, i]/mat.a[i, i];
        for k := i to ordem do
          mat.a[j, k] := mat.a[j, k] - aux*mat.a[i, k];
      end;
    end;
  end;
  d := 1;
  for k := 1 to ordem do d := d*mat.a[k, k];
  if Odd(inversao) then d := -d;
  Det := d;
end;

{ --------------------------------------------------------------------------- }

function Nulidade(mat: matriz; lambda: real; n: byte): byte;

{ ---> calcula a nulidade da matriz quadrada mat - lambda*I elevada  a  'n'   }

var
  LinhaNula: boolean;
  i, j, QuantLinhasNulas: byte;

begin
  if (mat.m <> mat.n) then Exit;
  for i := 1 to mat.m do mat.a[i, i] := mat.a[i, i] - lambda;
  ElevaMatriz(mat, n);
  EscalonaMatriz(mat, 2);
  QuantLinhasNulas := 0;
  for i := 1 to mat.m do
  begin
    LinhaNula := true;
    j := 0;
    repeat
      Inc(j);
      if (Abs(mat.a[i, j]) > erro3) then LinhaNula := false;
    until (j = mat.n) or (not LinhaNula);
    if LinhaNula then Inc(QuantLinhasNulas);
  end;
  Nulidade := QuantLinhasNulas;
end;

{ --------------------------------------------------------------------------- }

function Nul(mat: matriz): byte;

{ ---> calcula a nulidade da matriz mat   }

var
  LinhaNula: boolean;
  i, j, QuantLinhasNulas: byte;

begin
  EscalonaMatriz(mat, 2);
  QuantLinhasNulas := 0;
  for i := 1 to mat.m do
  begin
    LinhaNula := true;
    j := 0;
    repeat
      Inc(j);
      if (Abs(mat.a[i, j]) > erro3) then LinhaNula := false;
    until (j = mat.n) or (not LinhaNula);
    if LinhaNula then Inc(QuantLinhasNulas);
  end;
  Nul := QuantLinhasNulas;
end;

{ --------------------------------------------------------------------------- }

procedure MostraMAT(a: ind; inicio, fim, atual: byte);

var
  i: byte;

begin
  ClrScr;
  i := inicio;
  repeat
    Writeln('   ', a[i]);
    Inc(i);
  until (i > fim) or (i = inicio + 20);
  GoToXY(4, atual);
  Cor(0, 7); Write(a[inicio - 1 + atual]:8); Cor(7, 0);
end;

{ --------------------------------------------------------------------------- }

procedure LeIndicesMAT(var nome: string);

var
  a: ind;
  atual, total, i, inicio, InicioVirtual: byte;
  ch: char;
  arq: text;
  HouveErro: boolean;

begin
  Assign(arq, 'MAT');
  {$I-}
  Reset(arq);
  {$I+}
  if (IOResult <> 0) then
    Erro(3, 1, HouveErro)
  else
  begin
    Window(1, 1, 24, 24);
    Cor(7, 0);
    GoToXY(1, 1);
    Write(#218, #196, ' Matrizes no disco ', #196, #191);
    for i := 1 to 22 do
    begin
      GoToXY(1, i + 1); Write(#179, Spc(21), #179);
    end;
    Writeln;
    Write(#192, #196, ' Use ', #24, ' ', #25, ' ENTER ESC ');
    Write(#196, #217);
    Window(3, 3, 22, 23);
    i := 1; total := 0;
    repeat
      Readln(arq, a[i]);
      Inc(i); Inc(total);
    until Eof(arq) or (i > 100);
    Close(arq);
    inicio := 1;
    MostraMAT(a, inicio, total, inicio);
    repeat
      ch  := ReadKey;
      if (ch = #0) then
      begin
        ch := ReadKey;
        case ch of
          SetaPCima: if (inicio > 1) then Dec(inicio);
          SetaPBaixo: if (inicio < total ) then Inc(inicio);
        end;
        if (inicio <= 20) then
        begin
          atual := inicio;
          MostraMAT(a, 1, total, atual);
        end
        else
        begin
          atual := 20; InicioVirtual := inicio - 19;
          MostraMat(a, InicioVirtual, total, atual);
        end;
      end;
      if (ch = ENTER) then
      begin
        nome := Copy(a[inicio], 1, 8);
        Window(42, 17, 70, 19);
        GoToXY(5, 2);
      end
      else
        if (ch = ESC) then
        begin
          nome := '';
          Window(42, 17, 70, 19);
          GoToXY(5, 2);
        end;
    until ((ch = ESC) or (ch = ENTER));
  end;
end;

{ --------------------------------------------------------------------------- }

procedure LeMatrizDisco(var mat: matriz; nome: string; var HErro: boolean);

var
  tela: TelaPtr;
  arq: file of matriz;
  tipo, xpos, ypos, i, j, ponto: byte;
  texto: text;
  asterisco: boolean;

begin
  HErro := false;
  if (nome = '') then tipo := 0 else tipo := 1;
  New(tela);
  if (tipo = 0) then
  begin
    xpos := WhereX; ypos := WhereY;
    Write('Nome da matriz? ');
    Readln(nome);
  end;
  repeat
    asterisco := (Pos('*', nome) > 0) or (Pos('?', nome) > 0);
    if asterisco then
    begin
      SalvaTela(tela);
      LeIndicesMAT(nome);
      if (nome = '') then
      begin
        RestauraTela(tela);
        Janela(42, 17, ' Nome da Matriz? ', nome);
      end;
    end;
  until (not asterisco);
  Dispose(tela);
  if (tipo = 0) then Cls;
  ponto := Pos('.', nome);
  if (ponto > 0) then nome := Copy(nome, 1, ponto - 1);
  nome := nome + '.MAT';
  Assign(arq, nome);
  {$I-}
  Reset(arq);
  {$I+}
  if (IOResult = 0) then
  begin
    if (tipo = 0) then
    begin
      GoToXY(xpos, ypos); ClrEol;
    end;
    if (tipo = 0) then
      Write('Lendo a matriz do disco...')
    else
      Write('Lendo a matriz...');
    Read(arq, mat);
    Close(arq);
    if (tipo = 0) then
    begin
      GoToXY(xpos, ypos); ClrEol;
    end;
  end
  else
    if (mat.m <= 5) then
      Erro(2, 2, HErro)
    else
      Erro(2, 1, HErro);
end;

{ --------------------------------------------------------------------------- }

procedure SalvaMatriz(mat: matriz; op: byte; nome: string; tela: TelaPtr);

var
  arq: file of matriz;
  xpos, ypos, tipo, tam1, tam2, i, j, simb, quant: byte;
  texto: text;
  HouveErro: boolean;

begin
  if (nome = '') then
  begin
    tipo := 0;
    xpos := WhereX; ypos := WhereY;
    Write('Nome da matriz? ');
    Readln(nome);
  end
  else
    tipo := 1;
  simb := Pos('.', nome);
  if (simb > 0) then nome := Copy(nome, 1, simb - 1);
  nome := nome + '.MAT';
  Assign(arq, nome);
  {$I-}
  Rewrite(arq);
  {$I+}
  if (IOResult = 0) then
  begin
    if (tipo = 0) then
    begin
      GoToXY(xpos, ypos); ClrEol;
    end;
    Write('Salvando...');
    Write(arq, mat);
    Close(arq);
    if (tipo = 0) then
    begin
      GoToXY(xpos, ypos); ClrEol;
    end;
  end
  else
  begin
    RestauraTela(tela);
    if (op = 2) then
      Erro(1, 1, HouveErro)
    else
    begin
      if (op > 0) and (mat.m <= 5) then
        Erro(1, 2, HouveErro)
      else
        Erro(1, 1, HouveErro);
    end;
    Exit;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure Moldura(mat: matriz; imp: InfoPress; caso: byte; tela: TelaPtr);

var
  m, NStr: string;

begin
  if (caso = 5) then
  begin
    Str(imp.n, NStr); m := mens1[caso];
    mens1[caso] := mens1[caso] + NStr + ' ';
  end;
  MostraTela(tela);
  if (caso <> 0) then
  begin
    Window(10, 5, 71, 21); Cor(1, 7);
    GoToXY(Trunc((62 - Length(mens1[caso]))/2), 9);
    Write(mens1[caso]);
  end;
  if (caso = 5) then mens1[caso] := m;
  Window(11, 6, 69, 20); Cor(2, 0);
  MostraMatriz(mat, imp);
  Window(10, 22, 71, 22); Cor(7, 0);
  Write(mens2[caso, 2]);
  Window(11, 14, 69, 20); Cor(2, 7);
end;

{ --------------------------------------------------------------------------- }

procedure LeTeclado(var mat: matriz);

begin
  Cor(1, 7);
  Window(5, 8, 70, 22);
  LeMatriz(mat, 3);
end;

{ --------------------------------------------------------------------------- }

function LI(mat: matriz): boolean;

begin
  LI := (Nul(mat) = 0);
end;

{ --------------------------------------------------------------------------- }

procedure MatrizTransposta(var mat: matriz);

var
  aux: matriz;
  i, j: byte;

begin
  aux.m := mat.n; aux.n := mat.m;
  for i := 1 to aux.m do
    for j := 1 to aux.n do
      aux.a[i, j] := mat.a[j, i];
  mat := aux;
end;

{ --------------------------------------------------------------------------- }

procedure MatrizMudancaDeBase(mat1, mat2: matriz; var mat3: matriz;
                                                        var SaoBases: boolean);

var
  s: matriz;
  sol: resposta;
  i, j, k: byte; limat1, limat2: boolean;

begin
  mat3.m := mat1.m; mat3.n := mat1.n; SaoBases := true;
  if (not LI(mat1)) or (not LI(mat2)) or (mat1.m <> mat1.n)
                                                     or (mat2.m <> mat2.n) then
  begin
    Writeln('*** Vetores nao formam base ***');
    SaoBases := false;
    Exit;
  end;
  for k := 1 to mat1.m do
  begin
    s.m := mat1.m; s.n := mat1.n + 1;
    for i := 1 to s.m do
    begin
      for j := 1 to mat1.n do
        s.a[i, j]  := mat1.a[j, i];
      s.a[i, s.n] := mat2.a[k, i];
    end;
    ResolveSistema(s, sol);
    if sol.SistDet then
    begin
      for j := 1 to mat3.m do
        mat3.a[j, k] :=  sol.SolParticular[j];
    end
    else
    begin
      Write(' *** Erro ***');
      Exit;
    end;
  end;
end;

{ --------------------------------------------------------------------------- }

procedure MatrizRelBase(mat, b: matriz; var NMat: matriz);

{ ---> calcula a matriz NMat da transformacao mat com relacao 'a base b       }

var
  SaoBases, inv: boolean;
  i, j: byte;
  aux, MatInv: matriz;

begin
  NMat.m := mat.m; NMat.n := mat.n; aux.m := mat.m; aux.n := mat.n;
  for i := 1 to mat.m do
    for j :=  1 to mat.n do
      if (i = j) then aux.a[i, j]  := 1 else aux.a[i, j]  := 0;
  MatrizMudancaDeBase(aux, b, NMat, SaoBases);
  if not SaoBases then Halt;
  MatInv := NMat;
  InverteMatriz(MatInv, inv);
  if not inv then Halt;
  MultiplicaMatrizes(MatInv, mat, mat);
  MultiplicaMatrizes(mat, NMat, NMat);
end;

{ --------------------------------------------------------------------------- }

var
  arq1: file of real;
  arq2: file of byte;

begin
  Assign(arq1, 'MATRIZES.ERR');
  {$I-}
  Reset(arq1);
  {$I+}
  if (IOResult = 0) then
  begin
    Read(arq1, erro1, erro2, erro3);
    Close(arq1);
  end;
  Assign(arq2, 'ALC.DTA');
  {$I-}
  Reset(arq2);
  {$I+}
  if (IOResult = 0) then
  begin
    Read(arq2, decimais);
    Close(arq2);
  end;
end.

{ =========================================================================== }

